"""pyvizio utility constants."""

APK_SOURCE_PATH = "src"
RESOURCE_PATH = "resources/res/raw"
APP_NAMES_FILE = "apps.json"
APP_PAYLOADS_FILE = "apps_availability.json"
APP_NAMES_URL = "http://hometest.buddytv.netdna-cdn.com/appservice/vizio_apps_prod.json"
APP_PAYLOADS_URL = (
    "http://hometest.buddytv.netdna-cdn.com/appservice/app_availability_prod.json"
)
